/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Matt Eagleson (e.Messiah@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for mo0re details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <gnome.h>
#include <gtk/gtk.h>
#include <pthread.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>

#include "../uulib/uudeview.h"
#include "../uulib/fptools.h"

#include "pan.h"

GSList *queue = NULL;
GSList *queue_last = NULL;
pthread_mutex_t qlock = PTHREAD_MUTEX_INITIALIZER;

pthread_mutex_t qcond_m = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t qcond = PTHREAD_COND_INITIALIZER;

void
queue_append (queue_data* qdata)
{
	GSList *l;

	l = g_slist_alloc();
	l->data = qdata;

	if (queue) {
		g_assert(queue_last);
		g_slist_concat(queue_last, l);
		queue_last = l;
	} else
		queue = queue_last = l;
}


/*--------------------------------------------------------------------
 * put an item at the front of the queue
 *--------------------------------------------------------------------*/
void
queue_prepend (queue_data* qdata)
{
	if (queue_last)
		queue = g_slist_prepend(queue, qdata);
	else
		queue_last = queue = g_slist_prepend(queue, qdata);
}


/*---[ queue_get_next ]-----------------------------------------------
 * return next item on queue, remove link to current one
 *--------------------------------------------------------------------*/
// FIXME: is the old queue item actually being deleted? (possible memory leak)
queue_data*
queue_get_next (void)
{
	queue_data *qdata;

	pthread_mutex_lock(&qlock);
	if (queue) {
		qdata = (queue_data *) queue->data;
		queue = g_slist_remove_link(queue, queue);
		if (queue == NULL)
			queue_last = NULL;
		pthread_mutex_unlock(&qlock);
		return qdata;
	} else {
		pthread_mutex_unlock(&qlock);
		return NULL;
	}
}


/*--------------------------------------------------------------------
 * remove all items from the queue
 *--------------------------------------------------------------------*/

void
queue_purge (void)
{
	GSList *p;
	queue_data *qdata;
	gint in_mem = 0;
	gchar **sd;

	pthread_mutex_lock(&qlock);
	for (p = queue; p; p = p->next) {
		/* FIXME: Remove the "queued" icon in the article_list */
		g_free(p->data);
	}
	g_slist_free(queue);
	queue = NULL;
	queue_last = NULL;
	pthread_mutex_unlock(&qlock);
}

static void queue_run (queue_data * qdata)
{
	g_assert(qdata);
	
	switch (qdata->flags & 0x003F) {
	case QUEUE_ARTICLE_DOWNLOAD:
		if (go_online_ask() == FALSE)
			connect_wait();
		article_download (qdata);
		break;
	case QUEUE_ARTICLE_DECODE:
		if (go_online_ask() == FALSE)
			connect_wait();
		article_decode (qdata);
		break;
	case QUEUE_ARTICLELIST_LOAD:
		articlelist_load (qdata);
		break;
	case QUEUE_ARTICLELIST_DOWNLOAD_ALL:
	case QUEUE_ARTICLELIST_DOWNLOAD_NEW:
	case QUEUE_ARTICLELIST_DOWNLOAD_SAMPLE:
		if (go_online_ask() == FALSE)
			connect_wait();
		articlelist_download (qdata);
		break;
	case QUEUE_GROUPLIST_DOWNLOAD:
		if (go_online_ask() == FALSE)
			connect_wait();
		grouplist_download ();
		break;
	case QUEUE_ARTICLE_POST:
		if (go_online_ask() == FALSE)
			connect_wait();
		article_post (qdata);
			break;
		case QUEUE_MULTIPART_DECODE:
		if (go_online_ask() == FALSE)
			connect_wait();
		multipart_decode (qdata);
		break;
	default:
		g_error("Reached default in queue_processing()\n");
		break;
	}
	/* Put stuff here you want to do after processing
	 * queue items */
	gui_clean();
}

void
queue_processing (void)
{
	struct timeval now;
	struct timespec timeout;
	int retcode;
	queue_data *qdata;
	time_t last_run, current_time;
	
	while (1) {
	
		pthread_mutex_lock(&qcond_m);
		gettimeofday(&now, NULL);
		timeout.tv_sec = now.tv_sec + 20;
		timeout.tv_nsec = now.tv_usec * 1000;
		retcode = 0;
	
		/* Wait for something to hit the queue */
		retcode = pthread_cond_timedwait(&qcond, &qcond_m, &timeout);
		if ((retcode == ETIMEDOUT) && (queue == NULL)) {
			/* Nothing hit the queue before timeout */
			if (is_online()) {
				time(&current_time);
				if ((current_time - last_run) > 300)
					/* We've been idle for 300 seconds, so go offline */
					go_offline();
				else
					/* Send HELP as a no-operation to keep the connection alive */
					nntp_send_noop();
			}
		} else {
			/* Something hit the queue, do it */
			pthread_mutex_unlock(&qcond_m);
			while (qdata = queue_get_next()) {
				queue_run(qdata);
			}
			/* Record the last time we ran queue */
			pthread_mutex_lock(&qcond_m);
			time(&last_run);
		}
		pthread_mutex_unlock(&qcond_m);
	}
}


void
queue_init (void)
{
	/* Fire up the queue thread */
	pthread_create(&Pan.thread, NULL, (void *) &queue_processing, NULL);
	pthread_detach(Pan.thread);
}


void
queue_add_post(gchar* message)
{
	queue_data *qdata;

	qdata = (queue_data *) g_malloc0(sizeof(queue_data));
	qdata->str = message;
	qdata->flags = QUEUE_ARTICLE_POST;
	pthread_mutex_lock(&qlock);
	queue_prepend(qdata);
	pthread_mutex_unlock(&qlock);

	/* Signal the addition */
	pthread_mutex_lock(&qcond_m);
	pthread_cond_broadcast(&qcond);
	pthread_mutex_unlock(&qcond_m);
}


void
queue_add_grouplist_download (void)
{
	queue_data *qdata;

	qdata = (queue_data *) g_malloc0(sizeof(queue_data));
	qdata->flags = QUEUE_GROUPLIST_DOWNLOAD;
	pthread_mutex_lock(&qlock);
	queue = g_slist_prepend(queue, qdata);
	pthread_mutex_unlock(&qlock);
	/* Signal the addition */
	pthread_mutex_lock(&qcond_m);
	pthread_cond_broadcast(&qcond);
	pthread_mutex_unlock(&qcond_m);
}


void
queue_add_article_download (article_data* adata, guint flags)
{
	queue_data *qdata;
	GSList *temp_list;

	qdata = (queue_data *) g_malloc0(sizeof(queue_data));
	qdata->gdata = adata->group;
	qdata->str = g_strdup(adata->message_id);
	qdata->flags = flags;
	temp_list = g_slist_alloc();
	temp_list->data = qdata;
	pthread_mutex_lock(&qlock);
	if (flags & QUEUE_FOREGROUND)
		queue_prepend(qdata);
	else if (flags & QUEUE_BACKGROUND)
		queue_append(qdata);
	else
		g_error("Unreachable condition reached\n");
	pthread_mutex_unlock(&qlock);
	/* Signal the addition */
	pthread_mutex_lock(&qcond_m);
	pthread_cond_broadcast(&qcond);
	pthread_mutex_unlock(&qcond_m);
}


void
queue_add_multipart_download (article_data* adata, guint flags)
{
	queue_data *qdata;
	article_data *adata0;
	GSList *l;
	gchar *tmp;

	g_assert(adata && adata->threads && flags);

	qdata = (queue_data *) g_malloc0(sizeof(queue_data));
	qdata->gdata = adata->group;

	qdata->str =
		g_strdup_printf("%d\n%s\n%s", adata->linecount, adata->subject,
						adata->message_id);
	for (l = adata->threads; l; l = l->next) {
		adata0 = l->data;
		if (adata0->part > 0) {
			tmp = qdata->str;
			qdata->str = g_strconcat(tmp, "\n", adata0->message_id, NULL);
			g_free(tmp);
		}
	}
	qdata->flags = flags;
	pthread_mutex_lock(&qlock);
	queue_append(qdata);
	pthread_mutex_unlock(&qlock);
	/* Signal the addition */
	pthread_mutex_lock(&qcond_m);
	pthread_cond_broadcast(&qcond);
	pthread_mutex_unlock(&qcond_m);
}


void
queue_add_articlelist_download (group_data* gdata, guint flags)
{
	queue_data *qdata;

	qdata = (queue_data *) g_malloc0(sizeof(queue_data));
	qdata->gdata = gdata;
	qdata->flags = flags;
	pthread_mutex_lock(&qlock);
	if (flags & QUEUE_FOREGROUND)
		queue_prepend(qdata);
	else if (flags & QUEUE_BACKGROUND)
		queue_append(qdata);
	else
		g_error("Unreachable condition reached\n");
	pthread_mutex_unlock(&qlock);
	/* Signal the addition */
	pthread_mutex_lock(&qcond_m);
	pthread_cond_broadcast(&qcond);
	pthread_mutex_unlock(&qcond_m);
}
